from calendar import timegm
from datetime import timedelta, datetime

from playhouse.test_utils import assert_query_count

from data.database import (Tag, ManifestLegacyImage, TagToRepositoryTag, TagManifestToManifest,
                           TagManifest, Manifest)
from data.model.oci.tag import (find_matching_tag, get_most_recent_tag, list_alive_tags,
                                get_legacy_images_for_tags, filter_to_alive_tags,
                                filter_to_visible_tags, list_repository_tag_history,
                                get_expired_tag, get_tag, delete_tag,
                                delete_tags_for_manifest, change_tag_expiration,
                                set_tag_expiration_for_manifest, retarget_tag,
                                create_temporary_tag)
from data.model.repository import get_repository, create_repository

from test.fixtures import *

@pytest.mark.parametrize('namespace_name, repo_name, tag_names, expected', [
  ('devtable', 'simple', ['latest'], 'latest'),
  ('devtable', 'simple', ['unknown', 'latest'], 'latest'),
  ('devtable', 'simple', ['unknown'], None),
])
def test_find_matching_tag(namespace_name, repo_name, tag_names, expected, initialized_db):
  repo = get_repository(namespace_name, repo_name)
  if expected is not None:
    with assert_query_count(1):
      found = find_matching_tag(repo, tag_names)

    assert found is not None
    assert found.name == expected
    assert not found.lifetime_end_ms
  else:
    with assert_query_count(1):
      assert find_matching_tag(repo, tag_names) is None


def test_get_most_recent_tag(initialized_db):
  repo = get_repository('outsideorg', 'coolrepo')

  with assert_query_count(1):
    assert get_most_recent_tag(repo).name == 'latest'


def test_get_most_recent_tag_empty_repo(initialized_db):
  empty_repo = create_repository('devtable', 'empty', None)

  with assert_query_count(1):
    assert get_most_recent_tag(empty_repo) is None


def test_list_alive_tags(initialized_db):
  found = False
  for tag in filter_to_visible_tags(filter_to_alive_tags(Tag.select())):
    tags = list_alive_tags(tag.repository)
    assert tag in tags

    with assert_query_count(1):
      legacy_images = get_legacy_images_for_tags(tags)

    for tag in tags:
      assert ManifestLegacyImage.get(manifest=tag.manifest).image == legacy_images[tag.id]

    found = True

  assert found

  # Ensure hidden tags cannot be listed.
  tag = Tag.get()
  tag.hidden = True
  tag.save()

  tags = list_alive_tags(tag.repository)
  assert tag not in tags


def test_get_tag(initialized_db):
  found = False
  for tag in filter_to_visible_tags(filter_to_alive_tags(Tag.select())):
    repo = tag.repository

    with assert_query_count(1):
      assert get_tag(repo, tag.name) == tag
    found = True

  assert found


@pytest.mark.parametrize('namespace_name, repo_name', [
  ('devtable', 'simple'),
  ('devtable', 'complex'),
])
def test_list_repository_tag_history(namespace_name, repo_name, initialized_db):
  repo = get_repository(namespace_name, repo_name)

  with assert_query_count(1):
    results, has_more = list_repository_tag_history(repo, 1, 100)

  assert results
  assert not has_more


def test_list_repository_tag_history_with_history(initialized_db):
  repo = get_repository('devtable', 'history')

  with assert_query_count(1):
    results, _ = list_repository_tag_history(repo, 1, 100)

  assert len(results) == 2
  assert results[0].lifetime_end_ms is None
  assert results[1].lifetime_end_ms is not None

  with assert_query_count(1):
    results, _ = list_repository_tag_history(repo, 1, 100, specific_tag_name='latest')

  assert len(results) == 2
  assert results[0].lifetime_end_ms is None
  assert results[1].lifetime_end_ms is not None

  with assert_query_count(1):
    results, _ = list_repository_tag_history(repo, 1, 100, specific_tag_name='foobar')

  assert len(results) == 0


def test_list_repository_tag_history_all_tags(initialized_db):
  for tag in Tag.select():
    repo = tag.repository
    with assert_query_count(1):
      results, _ = list_repository_tag_history(repo, 1, 1000)

    assert (tag in results) == (not tag.hidden)


@pytest.mark.parametrize('namespace_name, repo_name, tag_name, expected', [
  ('devtable', 'simple', 'latest', False),
  ('devtable', 'simple', 'unknown', False),
  ('devtable', 'complex', 'latest', False),

  ('devtable', 'history', 'latest', True),
])
def test_get_expired_tag(namespace_name, repo_name, tag_name, expected, initialized_db):
  repo = get_repository(namespace_name, repo_name)

  with assert_query_count(1):
    assert bool(get_expired_tag(repo, tag_name)) == expected


def test_delete_tag(initialized_db):
  found = False
  for tag in list(filter_to_visible_tags(filter_to_alive_tags(Tag.select()))):
    repo = tag.repository

    assert get_tag(repo, tag.name) == tag
    assert tag.lifetime_end_ms is None

    with assert_query_count(4):
      assert delete_tag(repo, tag.name) == tag

    assert get_tag(repo, tag.name) is None
    found = True

  assert found


def test_delete_tags_for_manifest(initialized_db):
  for tag in list(filter_to_visible_tags(filter_to_alive_tags(Tag.select()))):
    repo = tag.repository
    assert get_tag(repo, tag.name) == tag

    with assert_query_count(5):
      assert delete_tags_for_manifest(tag.manifest) == [tag]

    assert get_tag(repo, tag.name) is None


@pytest.mark.parametrize('timedelta, expected_timedelta', [
  pytest.param(timedelta(seconds=1), timedelta(hours=1), id='less than minimum'),
  pytest.param(timedelta(weeks=300), timedelta(weeks=104), id='more than maxium'),
  pytest.param(timedelta(weeks=1), timedelta(weeks=1), id='within range'),
])
def test_change_tag_expiration(timedelta, expected_timedelta, initialized_db):
  now = datetime.utcnow()
  now_ms = timegm(now.utctimetuple()) * 1000

  tag = Tag.get()
  tag.lifetime_start_ms = now_ms
  tag.save()

  original_end_ms, okay = change_tag_expiration(tag, datetime.utcnow() + timedelta)
  assert okay
  assert original_end_ms == tag.lifetime_end_ms

  updated_tag = Tag.get(id=tag.id)
  offset = expected_timedelta.total_seconds() * 1000
  expected_ms = (updated_tag.lifetime_start_ms + offset)
  assert updated_tag.lifetime_end_ms == expected_ms


def test_set_tag_expiration_for_manifest(initialized_db):
  tag = Tag.get()
  manifest = tag.manifest
  assert manifest is not None

  set_tag_expiration_for_manifest(manifest, datetime.utcnow() + timedelta(weeks=1))

  updated_tag = Tag.get(id=tag.id)
  assert updated_tag.lifetime_end_ms is not None


def test_create_temporary_tag(initialized_db):
  tag = Tag.get()
  manifest = tag.manifest
  assert manifest is not None

  created = create_temporary_tag(manifest, 30)
  assert created is not None

  assert created.hidden
  assert created.name.startswith('$temp-')
  assert created.manifest == manifest
  assert created.lifetime_end_ms is not None
  assert created.lifetime_end_ms == (created.lifetime_start_ms + 30000)

  # Verify old-style tables.
  repository_tag = TagToRepositoryTag.get(tag=created).repository_tag
  assert repository_tag.lifetime_start_ts == int(created.lifetime_start_ms / 1000)
  assert repository_tag.lifetime_end_ts == int(created.lifetime_end_ms / 1000)
  assert repository_tag.name == created.name
  assert repository_tag.hidden

  tag_manifest = TagManifest.get(tag=repository_tag)
  assert TagManifestToManifest.get(tag_manifest=tag_manifest).manifest == manifest


def test_retarget_tag(initialized_db):
  repo = get_repository('devtable', 'history')
  results, _ = list_repository_tag_history(repo, 1, 100, specific_tag_name='latest')

  assert len(results) == 2
  assert results[0].lifetime_end_ms is None
  assert results[1].lifetime_end_ms is not None

  # Revert back to the original manifest.
  created = retarget_tag('latest', results[0].manifest, is_reversion=True,
                         now_ms=results[1].lifetime_end_ms + 10000)
  assert created.lifetime_end_ms is None
  assert created.reversion
  assert created.name == 'latest'
  assert created.manifest == results[0].manifest

  # Verify in the history.
  results, _ = list_repository_tag_history(repo, 1, 100, specific_tag_name='latest')

  assert len(results) == 3
  assert results[0].lifetime_end_ms is None
  assert results[1].lifetime_end_ms is not None
  assert results[2].lifetime_end_ms is not None

  assert results[0] == created

  # Verify old-style tables.
  repository_tag = TagToRepositoryTag.get(tag=created).repository_tag
  assert repository_tag.lifetime_start_ts == int(created.lifetime_start_ms / 1000)

  tag_manifest = TagManifest.get(tag=repository_tag)
  assert TagManifestToManifest.get(tag_manifest=tag_manifest).manifest == created.manifest


def test_retarget_tag_wrong_name(initialized_db):
  repo = get_repository('devtable', 'history')
  results, _ = list_repository_tag_history(repo, 1, 100, specific_tag_name='latest')
  assert len(results) == 2

  created = retarget_tag('someothername', results[1].manifest, is_reversion=True)
  assert created is None

  results, _ = list_repository_tag_history(repo, 1, 100, specific_tag_name='latest')
  assert len(results) == 2
